-- Henge Button handler

local AceOO = AceLibrary("AceOO-2.0")
local L = AceLibrary("AceLocale-2.2"):new("Henge")
local Tablet = AceLibrary("Tablet-2.0")

local buttonTable = {}
local mouseButtons = {
	[1] = "LeftButton",
	[2] = "RightButton",
	[3] = "MiddleButton"
}

HengeButton = AceOO.Class()
HengeButton.virtual = true

local HengeButtonAction = AceOO.Class(HengeButton)
local HengeButtonMenu = AceOO.Class(HengeButton)
local HengeButtonMenuChild = AceOO.Class()

		---------------------------------
		-- HengeButton local functions --
		---------------------------------

local function tabletPoint(frame)
	if frame:GetTop() > GetScreenHeight() / 2 then
		local x = frame:GetCenter()
		if x < GetScreenWidth() / 3 then
			return "TOPLEFT", "BOTTOMLEFT"
		elseif x < GetScreenWidth() * 2 / 3 then
			return "TOP", "BOTTOM"
		else
			return "TOPRIGHT", "BOTTOMRIGHT"
		end
	else
		local x = frame:GetCenter()
		if x < GetScreenWidth() / 3 then
			return "BOTTOMLEFT", "TOPLEFT"
		elseif x < GetScreenWidth() * 2 / 3 then
			return "BOTTOM", "TOP"
		else
			return "BOTTOMRIGHT", "TOPRIGHT"
		end
	end
end

		--------------------------------
		-- HengeButton main functions --
		--------------------------------

--[[ HengeButton:Add(arg) - Adds a new button, using the data from 'arg'.
		'arg' - Must be a _HengeButton definition table_
]]--
function HengeButton:Add(arg)
	if type(arg) ~= "table" or
	type(arg.id) ~= "string" or
	type(arg.options) ~= "table" or
	type(arg.type) ~= "string" or 
	type(arg.name) ~= "string" or 
	(arg.type ~= "action" and arg.type ~= "menu") then
		Henge:Debug("%s:Add() - Invalid argument.", self)
		return
	end
	if arg.type == "action" then
		if type(arg.actions) ~= "table" then
			Henge:Debug(
				"%s:Add() - Invalid HengeButtonAction definition.",
				self
			)
			return
		end
		buttonTable[arg.id] = HengeButtonAction:new(arg)
	elseif arg.type == "menu" then
		if type(arg.children) ~= "table" or type(arg.desc) ~= "string" then
			Henge:Debug(
				"%s:Add() - Invalid HengeButtonMenu definition.",
				self
			)
			return
		end
		buttonTable[arg.id] = HengeButtonMenu:new(arg)
	end
end

--[[ HengeButton:Get(id) - Gets a button specified by 'key'.
		'id' - Must be a number identifier for an existing button
]]--
function HengeButton:Get(id)
	if type(id) ~= "string" then
		Henge:Debug("%s:Get() - Invalid argument.", self)
		return nil
	elseif not AceOO.inherits(buttonTable[id], HengeButton) then
		Henge:Debug("%s:Get() - Button '%s' is invalid.", self, id)
		return nil
	end
	return buttonTable[id]
end

--[[ HengeButton:UpdateAll() - Updates all existing buttons.
]]--
function HengeButton:UpdateAll()
	for k, v in pairs(buttonTable) do
		v:Update()
	end
end

		---------------------------------------------
		-- HengeButton - virtual class for buttons --
		---------------------------------------------

function HengeButton:ToString()
	return "HengeButton"
end

function HengeButton.prototype:init(arg)
	HengeButton.super.prototype.init(self)
	self.id = arg.id
	self.name = arg.name
	self.desc = arg.desc
	self.icon = arg.icon
	self.OnUpdateHook = arg.hook
	self.db = Henge:AcquireDBNamespace("Button"..self.id)
	Henge:RegisterDefaults("Button"..self.id, "char", arg.options)
	if type(arg.counter) == "string" then
		self.counter = HengeCounter:Get(arg.counter)
		HengeCounter:AddWatch("HengeButton"..self.id, function(count, always)
			self:OnCounterUpdate(count, always)
		end)
	end
	self.dragging = false
end

function HengeButton.prototype:OnTooltipUpdate()
	Tablet:SetTitle(self.name)
	local cat = Tablet:AddCategory(
		'columns', 1,
		'child_textR', '1',
		'child_textG', '1',
		'child_textB', '1'
	)
	cat:AddLine('text', L["Empty button."])
end

function HengeButton.prototype:GetAnchor()
	local aos = self:GetOption("anchor")
	local anchor, offset = nil, 0
	if type(aos) == "string" then
		local sep = strfind(aos, ":")
		local a1, a2 = aos, nil
		if sep then
			a1 = strlower(strsub(aos, 1, sep - 1))
			a2 = strsub(aos, sep + 1)
		end
		if a1 == "Minimap" then
			anchor = Minimap
			offset = 80
		elseif a1 == "button" then
			local button = HengeButton:Get(a2)
			if AceOO.inherits(button, HengeButton) then
				anchor = button.frame
				offset = 32
			else
				anchor = Minimap
				offset = 80
			end
		elseif a1 == "sphere" then
			local sphere = HengeSphere:Get(a2)
			if AceOO.inherits(sphere, HengeSphere) then
				anchor = sphere.frame
				offset = 56
			else
				anchor = Minimap
				offset = 80
			end
		end
	end
	return anchor, offset
end

--[[ <HengeButton instance>:Position(x, y) - Unanchors the button if it was
	anchored, and moves it to the screen location specified by x and y.
		'x', 'y' - Must be numeric screen coordinates
]]-- ===OR===
--[[ <HengeButton instance>:Position(anchor, angle) - Anchors the button to
	'anchor', and moves it to the diameter of 'anchor' at 'angle' degrees.
		'anchor' - Must be a _HengeButton anchor string_
		'angle' - Must be a valid geometric angle
]]--
function HengeButton.prototype:Position(arg1, arg2)
	if type(arg1) == "string" and type(arg2) == "number" then
		self:SetOption("anchor", arg1)
		self:SetOption("angle", arg2)
		self:UpdatePosition()
	elseif type(arg1) == "number" and type(arg2) == "number" then
		self:SetOption("anchor", nil)
		self:SetOption("xpos", arg1)
		self:SetOption("ypos", arg2)
	else
		Henge:Debug("%s:Position() - Invalid argument.", self)
	end
end

--[[ <HengeButton instance>:UpdatePosition() - Moves the button to the
	location configured for it.
]]--
function HengeButton.prototype:UpdatePosition()
	local anchor, offset = self:GetAnchor()
	local x, y, angle = 0, 0, 0
	if anchor then
		local scale = anchor:GetEffectiveScale()
		angle = self:GetOption("angle")
		if not angle then angle = 10 end
		angle = math.rad(angle)
		x = (math.cos(angle) * offset) * scale
		y = (math.sin(angle) * offset) * scale
	else
		anchor = UIParent
		local scale = anchor:GetEffectiveScale()
		x = self:GetOption("xpos") * scale
		y = self:GetOption("ypos") * scale
	end
	self.frame:ClearAllPoints()
	self.frame:SetParent(anchor)
	self.frame:SetPoint("CENTER", anchor, "CENTER", x, y)
end

function HengeButton.prototype:GetIcon()
	local texpath = self.icon
	if (not texpath) then
		texpath = "Interface\\AddOns\\Henge\\icon"
	end
	return texpath
end

--[[ <HengeButton instance>:UpdateTexture() - Changes the button's normal
	texture to the one from its config.
]]--
function HengeButton.prototype:UpdateTexture()
	self.texture:SetTexture(self:GetIcon())
	self.texture:Show()
end

function HengeButton.prototype:Update()
	if type(self.UpdateHook) == "function" then
		self:UpdateHook()
	end
	self:UpdatePosition()
	self:UpdateTexture()
end

--[[ <HengeButton instance>:RefreshCounter(count, always) - Called by the
	OnUpdate function of a HengeCounter, to update the counter on the button.
		'count' - Must be a string or number representing the counter's value
		'always' - A false value causes the counter text to be hidden when
			the count is at 0, true shows the text even if it's at 0
]]--
function HengeButton.prototype:OnCounterUpdate(count, always)
	self.cdata.count = count
	self.cdata.always = always
	if self.cdata.always == true or self.cdata.count > 0.001 then
		self.text:SetText(tostring(self.cdata.count))
		self.text:Show()
	else
		self.text:SetText("")
		self.text:Hide()
	end
end

function HengeButton.prototype:OnDragStart(mbs)
	if mbs == "LeftButton" then
		if not Henge:IsLocked() then
			if type(self:GetOption("anchor")) ~= "string" then
				this:StartMoving()
			elseif self.frame:CanChangeProtectedState() then
				this:SetScript("OnUpdate", function() self:OnMove() end)
				self.dragging = true
			end
		end
	end
end

function HengeButton.prototype:OnDragStop()
	if type(self:GetOption("anchor")) ~= "string" then
		this:StopMovingOrSizing()
		local x, y = GetCenter()
		self:Position(x, y)
	elseif self.dragging then
		this:SetScript("OnUpdate", nil)
		self.dragging = false
	end
end

function HengeButton.prototype:OnMove()
	local astr = self:GetOption("anchor")
	local anchor, _ = self:GetAnchor()
	local scale = anchor:GetEffectiveScale()
	local x, y = GetCursorPosition()
	local sx, sy = anchor:GetCenter()
	x, y = (x / scale) - sx, (y / scale) - sy
	local angle = math.deg(math.atan2(y, x))
	self:Position(astr, angle)
end

function HengeButton.prototype:GetOption(opt)
	if (type(opt) ~= "string") then
		Henge:Debug("%s:GetOption() - Invalid option.", self)
		return nil
	end
	return self.db.char[opt]
end

function HengeButton.prototype:SetOption(opt, arg)
	if (type(opt) ~= "string") then
		Henge:Debug("%s:SetOption() - Invalid option.", self)
		return
	end
	if type(arg) == "function" then
		Henge:Debug("%s:SetOption() - Invalid argument.", self)
		return
	end
	self.db.char[opt] = arg
end

		----------------------------------------------------
		-- HengeButtonAction - handles the action buttons --
		----------------------------------------------------

function HengeButtonAction:ToString()
	return "HengeButtonAction"
end

function HengeButtonAction.prototype:init(arg)
	HengeButtonAction.super.prototype.init(self, arg)
	local globname = "HengeButton"..self.id
	self.frame = CreateFrame("Button", globname, nil,
		"HengeButtonTemplate,SecureActionButtonTemplate")
	self.frame.hbObj = self
	self.texture = getglobal(globname.."Texture")
	self.highlight = getglobal(globname.."Highlight")
	self.highlight:SetVertexColor(1.0, 0.5, 0.0)
	self.text = getglobal(globname.."Text")
	self.action = {}
	for mbs, act in pairs(arg.actions) do
		self.action[mbs] = HengeAction:Get(act)
		self.action[mbs]:Apply(self.frame, mbs)
	end
	Tablet:Register(self.frame,
		'children', function() self:OnTooltipUpdate() end,
		'point', tabletPoint,
		'data', {}
	)
	self:Update()
	self.frame:Show()
end

function HengeButtonAction.prototype:OnTooltipUpdate()
	Tablet:SetTitle(self.name)
	local cat = Tablet:AddCategory(
		'columns', 2,
		'child_textR', 1,
		'child_textG', 1,
		'child_textB', 0,
		'child_text2R', 1,
		'child_text2G', 1,
		'child_text2B', 1
	)
	for mbs, action in pairs(self.action) do
		local an, ad, ai = action:ActionInfo()
		cat:AddLine('text', mbs, 'text2', an)
	end
end

function HengeButtonAction.prototype:GetIcon()
	local texpath = self.icon
	if (not texpath) then
		texpath = "action:LeftButton"
	end
	if texpath then
		local tsep = strfind(texpath, ":")
		if tsep then
			local tty = strsub(texpath, 1, tsep - 1)
			local tid = strsub(texpath, tsep + 1)
			if tty == "action" then
				if AceOO.inherits(self.action[tid], HengeAction) then
					local _, _, icon = self.action[tid]:ActionInfo()
					texpath = icon
				end
			end
		end
	end
	return texpath
end

		------------------------------------------------
		-- HengeButtonMenu - handles the menu buttons --
		------------------------------------------------

function HengeButtonMenu:ToString()
	return "HengeButtonMenu"
end

function HengeButtonMenu.prototype:init(arg)
	HengeButtonMenu.super.prototype.init(self, arg)
	self.frame = CreateFrame(
		"Button", "HengeButton"..self.id, nil,
		"HengeButtonTemplate,SecureActionButtonTemplate,"..
		"SecureAnchorUpDownTemplate"
	)
	self.header = CreateFrame(
		"Frame", nil, self.frame, "SecureStateHeaderTemplate"
	)
	self.frame:SetAttribute("*anchorchild*", self.header)
	self.frame:SetAttribute("*childraise*", true)
	self.frame:SetAttribute("*onmouseupbutton2", "rightclick")
	self.frame:SetAttribute("*childstate-rightclick", "^rightclick")
	self.frame:SetAttribute("*childofsx*", 1)
	self.frame:SetAttribute("*childofsy*", 1)
	self.header:SetAttribute("statemap-anchor-rightclick", "0-1")
	self.header:SetAttribute("delaystatemap-anchor-rightclick", "0:0")
	self.header:SetAttribute("delaytimemap-anchor-rightclick", "0:1")
	self.header:SetAttribute("delayhovermap-anchor-rightclick", "0:true")
	self.frame.hbObj = self
	self.texture = getglobal("HengeButton"..self.id.."Texture")
	self.highlight = getglobal("HengeButton"..self.id.."Highlight")
	self.highlight:SetVertexColor(1.0, 0.5, 0.0)
	self.text = getglobal("HengeButton"..self.id.."Text")
	local scale = self.frame:GetEffectiveScale()
	self.child = {}
	for key, arg in ipairs(arg.children) do
		self.child[arg.id] = HengeButtonMenuChild:new(self, key, arg, scale)
		self.header:SetAttribute("addchild", self.child[arg.id].frame)
	end
	self:SetAction(self:GetOption("action"))
	Tablet:Register(self.frame,
		'children', function() self:OnTooltipUpdate() end,
		'point', tabletPoint,
		'data', {}
	)
	self:Update()
	self.frame:Show()
end

function HengeButtonMenu.prototype:OnTooltipUpdate()
	Tablet:SetTitle(self.name)
	local cat = Tablet:AddCategory(
		'columns', 2,
		'child_textR', 1,
		'child_textG', 1,
		'child_textB', 0,
		'child_text2R', 1,
		'child_text2G', 1,
		'child_text2B', 1
	)
	if AceOO.inherits(self.action, HengeAction) then
		local an, ad, ai = self.action:ActionInfo()
		cat:AddLine('text', "LeftClick", 'text2', an)
	end
	cat:AddLine('text', "RightClick", 'text2', L["Show button menu."])
end

function HengeButtonMenu.prototype:GetIcon()
	local texpath = self.icon
	if (not texpath) then
		texpath = "Interface\\AddOns\\Henge\\icon"
	elseif texpath == "action" then
		if AceOO.inherits(self.action, HengeAction) then
			local _, _, icon = self.action:ActionInfo()
			texpath = icon
		else
			texpath = nil
		end
	else
		local tsep = strfind(texpath, ":")
		if tsep then
			local tty = strsub(texpath, 1, tsep - 1)
			local tid = strsub(texpath, tsep + 1)
			if tty == "child" then
				if AceOO.inherits(self.child[tid], HengeButtonMenuChild) then
					texpath = self.child[tid]:GetIcon()
				end
			end
		end
	end
	return texpath
end

function HengeButtonMenu.prototype:UpdatePosition()
	HengeButton.prototype.UpdatePosition(self)
	for name, child in pairs(self.child) do
		child:UpdatePosition()
	end
end

function HengeButtonMenu.prototype:UpdateTexture()
	HengeButton.prototype.UpdateTexture(self)
	for name, child in pairs(self.child) do
		child:UpdateTexture()
	end
end

function HengeButtonMenu.prototype:SetAction(action)
	if type(action) == "string" then
		self.action = HengeAction:Get(action)
	else
		self.action = action
	end
	if AceOO.inherits(self.action, HengeAction) then
		self.action:Apply(self.frame, "LeftButton")
		local an, _, _ = self.action:ActionInfo()
		self:SetOption("action", self.action.id)
	else
		self.action = HengeAction:Get("None")
		self:SetOption("action", nil)
	end
	self:UpdateTexture()
end

		------------------------------------------------------------
		-- HengeButtonMenuChild - handles the buttons on the menu --
		------------------------------------------------------------

function HengeButtonMenuChild:ToString()
	return "HengeButtonMenuChild"
end

function HengeButtonMenuChild.prototype:init(pb, key, arg, scale)
	HengeButtonMenuChild.super.prototype.init(self)
	self.parent = pb
	self.cnum = key
	self.id = arg.id
	self.name = arg.name
	self.desc = arg.desc
	self.icon = arg.icon
	local globname = "HengeButtonMenuChild"..self.id
	self.frame = CreateFrame("Button", globname, nil,
		"HengeButtonTemplate,SecureActionButtonTemplate")
	self.frame.hbObj = self
	self.texture = getglobal(globname.."Texture")
	self.highlight = getglobal(globname.."Highlight")
	self.highlight:SetVertexColor(1.0, 0.5, 0.0)
	self.text = getglobal(globname.."Text")
	self.action = HengeAction:Get(arg.action)
	if AceOO.inherits(self.action, HengeAction) then
		self.action:Apply(self.frame, "LeftButton")
	else
		self.action = HengeAction:Get("none")
	end
	self.frame:SetAttribute("type2", "menu")
	self.frame.menu = function() self:OnRightClick() end
	Tablet:Register(self.frame,
		'children', function() self:OnTooltipUpdate() end,
		'point', tabletPoint,
		'data', {}
	)
	if type(arg.counter) == "string" then
		self.counter = HengeCounter:Get(arg.counter)
		HengeCounter:AddWatch("HengeButton"..self.id, function(count, always)
			self:OnCounterUpdate(count, always)
		end)
	end
	self.frame:SetToplevel(true)
	self.frame:SetFrameStrata("DIALOG")
	self.frame:SetScale(0.90)
	self.frame:SetAttribute("showstates", "1")
	self.frame:SetAttribute("newstate", ";")
	self:UpdatePosition()
end

function HengeButtonMenuChild.prototype:GetIcon()
	if type(self.icon) ~= "string" or self.icon == "action" then
		if AceOO.inherits(self.action, HengeAction) then
			local _, _, icon = self.action:ActionInfo()
			return icon
		else
			return nil
		end
	end
	return self.icon
end

function HengeButtonMenuChild.prototype:OnTooltipUpdate()
	Tablet:SetTitle(self.name)
	local cat = Tablet:AddCategory(
		'columns', 2,
		'child_textR', 1,
		'child_textG', 1,
		'child_textB', 0,
		'child_text2R', 1,
		'child_text2G', 1,
		'child_text2B', 1
	)
	local an, ad, ai = self.action:ActionInfo()
	cat:AddLine('text', "LeftButton", 'text2', an)
	cat:AddLine('text', "RightButton", 'text2', "Set as menu default.")
end

function HengeButtonMenuChild.prototype:OnRightClick()
	if self.parent.frame:CanChangeProtectedState() then
		self.parent:SetAction(self.action)
	end
end

function HengeButtonMenuChild.prototype:UpdatePosition()
	if not self.frame:CanChangeProtectedState() then
		return
	end
	local scale = self.frame:GetEffectiveScale()
	local angle = 90
	if type(self.parent:GetOption("child_angle")) == "number" then
		angle = self.parent:GetOption("child_angle") or 90
	elseif type(self.parent:GetOption("anchor")) == "string" then
		angle = self.parent:GetOption("angle") or 90
	end
	angle = math.rad(angle)
	local ofsx = (math.cos(angle) * 28 * self.cnum) * scale
	local ofsy = (math.sin(angle) * 28 * self.cnum) * scale
	self.frame:SetAttribute("ofsx", ofsx)
	self.frame:SetAttribute("ofsy", ofsy)
end

HengeButtonMenuChild.prototype.UpdateTexture = 
	HengeButton.prototype.UpdateTexture

HengeButtonMenuChild.prototype.OnCounterUpdate =
	HengeButton.prototype.OnCounterUpdate

